/*
 * Copyright (c) 2009 Lukas Mejdrech
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup packet
 *  @{
 */

/** @file
 *  Packet client implementation.
 */

#include <async.h>
#include <errno.h>
#include <unistd.h>
//#include <stdio.h>
#include <string.h>

#include <ipc/ipc.h>
#include <sys/mman.h>

#include "../../err.h"
#include "../../messages.h"

#include "packet.h"
#include "packet_header.h"
#include "packet_client.h"

/** Obtains the packet from the packet server as the shared memory block.
 *  Creates the local packet mapping as well.
 *  @param phone The packet server module phone. Input parameter.
 *  @param packet The packet reference pointer to store the received packet reference. Output parameter.
 *  @param packet_id The packet identifier. Input parameter.
 *  @param size The packet total size in bytes. Input parameter.
 *  @returns EOK on success.
 *  \todo ipc_share_in_start() error?
 *  @returns Other error codes as defined for the pm_add() function.
 */
int packet_return( int phone, packet_ref packet, packet_id_t packet_id, size_t size );

int packet_copy_data( packet_t packet, const void * data, size_t length ){
	if( ! packet_is_valid( packet )) return EINVAL;
	if( packet->data_start + length >= packet->length ) return ENOMEM;
	memcpy(( void * ) packet + packet->data_start, data, length );
	if( packet->data_start + length > packet->data_end ){
		packet->data_end = packet->data_start + length;
	}
	return EOK;
}

void * packet_prefix( packet_t packet, size_t length ){
	if(( ! packet_is_valid( packet )) || ( packet->data_start - sizeof( struct packet ) - 2 * packet->addr_len < length )) return NULL;
	packet->data_start -= length;
	return ( void * ) packet + packet->data_start;
}

void * packet_suffix( packet_t packet, size_t length ){
	if(( ! packet_is_valid( packet )) || ( packet->data_end + length >= packet->length )) return NULL;
	packet->data_end += length;
	return ( void * ) packet + packet->data_end - length;
}

int packet_trim( packet_t packet, size_t prefix, size_t suffix ){
	if( ! packet_is_valid( packet )) return EINVAL;
	if( prefix + suffix > packet->data_end - packet->data_start ) return ENOMEM;
	packet->data_start += prefix;
	packet->data_end -= suffix;
	return EOK;
}

packet_id_t packet_get_id( const packet_t packet ){
	return packet_is_valid( packet ) ? packet->packet_id : 0;
}

int packet_get_addr( const packet_t packet, uint8_t ** src, uint8_t ** dest ){
	if( !( packet_is_valid( packet ) && src && dest )) return EINVAL;
	if( ! packet->addr_len ) return 0;
	* src = ( void * ) packet + packet->src_addr;
	* dest = ( void * ) packet + packet->dest_addr;
	return packet->addr_len;
}

size_t packet_get_data_length( const packet_t packet ){
	if( ! packet_is_valid( packet )) return 0;
	return packet->data_end - packet->data_start;
}

void * packet_get_data( const packet_t packet ){
	if( ! packet_is_valid( packet )) return NULL;
	return ( void * ) packet + packet->data_start;
}

int packet_set_addr( packet_t packet, const uint8_t * src, const uint8_t * dest, size_t addr_len ){
	size_t	padding;

	if( ! packet_is_valid( packet )) return EINVAL;
	if( packet->addr_len >= addr_len ) return ENOMEM;
	padding = packet->addr_len - addr_len;
	if( src ){
		memcpy(( void * ) packet + packet->src_addr, src, addr_len );
		memset(( void * ) packet + packet->src_addr + addr_len, 0, padding );
	}else{
		memset(( void * ) packet + packet->src_addr + addr_len, 0, packet->addr_len );
	}
	if( dest ){
		memcpy(( void * ) packet + packet->dest_addr, dest, addr_len );
		memset(( void * ) packet + packet->dest_addr + addr_len, 0, padding );
	}else{
		memset(( void * ) packet + packet->dest_addr + addr_len, 0, packet->addr_len );
	}
	return EOK;
}

int packet_translate( int phone, packet_ref packet, packet_id_t packet_id ){
	ERROR_DECLARE;

	unsigned int		size;

	if( ! packet ) return EINVAL;
	* packet = pm_find( packet_id );
	if( * packet ) return EOK;
	ERROR_PROPAGATE( async_req_1_1( phone, NET_PACKET_GET_SIZE, packet_id, & size ));
	return packet_return( phone, packet, packet_id, size );
}

int packet_return( int phone, packet_ref packet, packet_id_t packet_id, size_t size ){
	ERROR_DECLARE;

	aid_t		message;
	ipc_call_t	answer;
	ipcarg_t	result;

	message = async_send_1( phone, NET_PACKET_GET, packet_id, & answer );
	* packet = ( packet_t ) as_get_mappable_page( size );
	if( ERROR_OCCURRED( ipc_share_in_start_0_0( phone, * packet, size ))
	|| ERROR_OCCURRED( pm_add( * packet ))){
		munmap( * packet, size );
		async_wait_for( message, NULL );
		return ERROR_CODE;
	}
	async_wait_for( message, & result );
	return result;
}

packet_t packet_get_4( int phone, size_t max_content, size_t addr_len, size_t max_prefix, size_t max_suffix ){
	ERROR_DECLARE;

	packet_id_t	packet_id;
	unsigned int size;
	packet_t packet;

	if( ERROR_OCCURRED( async_req_4_2( phone, NET_PACKET_CREATE_4, max_content, addr_len, max_prefix, max_suffix, & packet_id, & size ))
	|| ERROR_OCCURRED( packet_return( phone, & packet, packet_id, size ))){
		return NULL;
	}
	return packet;
}

packet_t packet_get_1( int phone, size_t content ){
	ERROR_DECLARE;

	packet_id_t	packet_id;
	unsigned int	size;
	packet_t	packet;

	if( ERROR_OCCURRED( async_req_1_2( phone, NET_PACKET_CREATE_1, content, & packet_id, & size ))
	|| ERROR_OCCURRED( packet_return( phone, & packet, packet_id, size ))){
		return NULL;
	}
	return packet;
}

void packet_release( int phone, packet_id_t packet_id ){
	async_msg_1( phone, NET_PACKET_RELEASE, packet_id );
}

/** @}
 */
